;;;
;;;  The arc prototype
;;;

;;; $Revision: 1.1 $

(defproto arc '(datasets num graph-number menu menu-items1 menu-items2
                  current-dataset settings) () )
(send arc :slot-value 'num 0)
(send arc :slot-value 'graph-number 0)
(send arc :add-slot 'lc-number 0)

(defmeth arc :delete-dataset (dataset)
"Method args: (dataset)
Deletes dataset from the list of datasets."
  (setf (slot-value 'datasets) (remove dataset (slot-value 'datasets)))
  (send self :update-menu)
; (when *arc-packages*
;  (unintern (string-to-symbol (string-upcase (send dataset :name))) 
;           (find-package "USER")))
  (let ((d (first (last (send self :datasets)))))
  (when d
        (send self :set-menus d)
        (send self :use-dataset (send d :name)))))

(defmeth arc :datasets (&optional dataset)
"Method args: (&optional dataset)
Returns the list of datasets, and optionally adds dataset to the list."
    (when dataset (setf (slot-value 'datasets) 
                      (append (slot-value 'datasets) (list dataset)))
                  (send self :update-menu)
                  (send dataset :add-slot 'dataset-number
                      (send self :dataset-number)))
    (slot-value 'datasets))

(defmeth arc :names ()
"Method args: ()
Returns the names of all datasets."
    (mapcar #'(lambda (d) (send d :name)) (send self :datasets)))

(defmeth arc :last-dataset () 
"Method args: ()
Returns the most recent dataset."
(first (last (send self :datasets))))

(defmeth arc :dataset-number ()
"Method args: ()
Increments the slot 'num by one and returns the value.  Used to number
datasets."
  (setf (slot-value 'num) (1+ (slot-value 'num)))
  (slot-value 'num))

(defmeth arc :graph-number (&optional (update t))
"Method args: (&optional (update t))
Increments the graph number if update is t, and then returns the value."
  (if update (setf (slot-value 'graph-number) (1+ (slot-value 'graph-number))))
  (slot-value 'graph-number))

(defmeth arc :lc-number (&optional (update t))
"Method args: (&optional (update t))
Returns/updates the number of the next graphical regression model."
  (if update (setf (slot-value 'lc-number) (1+ (slot-value 'lc-number))))
  (slot-value 'lc-number))

(defmeth arc :menus ()
"Method args: ()
Returns a list of menus for all datasets."
 (let* ((d (send self :datasets)))
   (mapcar #'(lambda (s)
    (cons (send s :menu) (reverse
     (mapcar #'(lambda (m) (send m :menu)) (send s :models))))) d)))

(defmeth arc :order-menus ()
"Method args: ()
Installs dataset menus in the right order."
 (let* ((r (combine (send arc :menus))))
  (mapcar #'(lambda (a) (send a :remove)) r)
  (mapcar #'(lambda (a) (send a :install)) r)))

(defmeth arc :set-menus (dataset)
"Method args: (dataset)
Deletes all datasets and models from the menubar except for the selected
dataset and its models."
 (unless (eq dataset (slot-value 'current-dataset))
  (let ((cd (slot-value 'current-dataset)))
   (when cd (send (send cd :menu) :remove)
            (when *show-plot-for-one-dataset*
              (send cd :add-slot 'plot-locations
               (mapcar #'(lambda (p) (send p :location)) (send cd :graphs)))
              (mapcar #'(lambda (p) (send p :hide-window)) (send cd :graphs)))
            (mapcar #'(lambda (m) (send (send m :menu) :remove))
                  (send cd :models)))
   (send (send dataset :menu) :install)
   (setf (slot-value 'current-dataset) dataset)
   (when (and *show-plot-for-one-dataset*  
              (send dataset :has-slot 'plot-locations))
    (mapcar #'(lambda (p) (send p :show-window)) (send dataset :graphs))
    (mapcar #'(lambda (p l) (apply #'send p :location l))
      (send dataset :graphs) (send dataset :slot-value 'plot-locations))
    )
   (mapcar #'(lambda (m) (send (send m :menu) :install)) 
             (reverse (send dataset :models))))))

(defmeth arc :active-dataset () (send self :current-dataset))
(defmeth arc :current-dataset () (slot-value 'current-dataset))
 
(defmeth arc :find-dataset (name)
"Method args: (name)
Returns the dataset-proto with menubar name name, or nil if no such object."
	(let* ((d (send self :datasets))
          (n (mapcar #'(lambda (d) (send d :name)) d))
          (pos (position name n :test #'equal)))
   (when pos (select d pos))))

(defmeth arc :menu ()
"Method args: ()
Returns the arc menu."
 (let* ((menu (slot-value 'menu))
        (n *arc-name*))
  (if menu menu (setf (slot-value 'menu) 
     (send self :make-menu (ms n n (format nil "&~a" n)))))))


(defmeth arc :make-menu (title)
"Method args: (title)
Creates Arc menu with title title."
 (let* ((items1 (mapcar #'(lambda (a) (make-menu-item a self))
                  *arc-menu-items1*))
        (items2 (mapcar #'(lambda (a) (make-menu-item a self))
                  *arc-menu-items2*))
        (menu (send menu-proto :new title)))
  (apply #'send menu :append-items (append items1 items2))
  (setf (slot-value 'menu-items1) items1)
  (setf (slot-value 'menu-items2) items2)
  (send menu :install)
; (if (eq *computer-type* 'unix)
;     (send xlisp::*fake-menu-bar* :location 33 73))
  menu))


(defmeth arc :update-menu ()
"This method updates the arc menu to include datast names."
  (let* ((menu (send self :menu))
         (items1 (slot-value 'menu-items1))
         (items2 (slot-value 'menu-items2))
         (d1 (send dash-item-proto :new))
         (d2 (send dash-item-proto :new))
         (ds (mapcar #'(lambda (d) (send menu-item-proto :new d
                   :action #'(lambda () 
                       (send self :use-dataset d)
                       (send self :set-menus (send self :find-dataset d)))))
             (send arc :names))))
   (apply #'send menu :delete-items (send menu :items))
   (apply #'send menu :append-items (if ds (combine items1 d1 ds d2 items2)
           (append items1 items2)))
   menu))

(defmeth arc :use-dataset (name)
"Method args: (name)
Makes the dataset with name name the current dataset."
  (send (send self :find-dataset name) :intern-data)
  (send (send self :find-dataset name) :intern-models))

(defparameter *arc-menu-items1*
 (list 'load-file-menu-item))
(defparameter *arc-menu-items2* (list 
 'dribble-menu-item 'p-value-calculator 'quantile-calculator
 'about-menu-item 'help-menu-item 'settings-menu-item 'quit-menu-item))

(rc-menu-item 'load-file-menu-item 
;  (ms "Read a datafile" "Read a datafile" "&Read a datafile") :read-datafile)
   (ms "Load" "Load" "&Load") :read-datafile)
(defmeth arc :read-datafile ()
"Method args: ()
Gets a file name and attempts to load that file."
    (let ((filename (bjm-file-dialog "Load file..." :mode 'open)))
      (when filename (arc-load filename :add-suffix nil))))

(rc-menu-item 'dribble-menu-item 
  (ms "Dribble..." "Dribble..." "Dribble...") :dribble)
(defmeth arc :dribble () 
"Method args: ()
Open/close the dribble file."
 (let* ((i (send (send self :menu) :items))
        (d (select i (position "Dribble..." (mapcar #'(lambda (a) 
            (send a :title)) i) :test #'equal)))) 
   (case (send d :mark)
    (nil (let ((filename (bjm-file-dialog "Dribble file name:" :mode 'create)))
          (when filename
           (dribble filename)
           (send d :mark t))))
    (t (dribble) (send d :mark nil)))))

(rc-menu-item 'p-value-calculator 
  (ms "Calculate probability..." "Calculate probability..." 
      "Calculate &probability...")
  :pvalcalc)

(rc-menu-item 'quantile-calculator 
  (ms "Calculate quantile..." "Calculate quantile..." "Calculate \&quantile...")
  :cvalcalc)

(defmeth arc :pvalcalc () 
"Method args: ()
Calls the p-value-calculator function."
(p-value-calculator))

(defmeth arc :cvalcalc () 
"Method args: ()
Calls the p-value-calculator function."
(quantile-calculator))

(rc-menu-item 'about-menu-item 
  (ms (format nil "About ~a" *arc-name*)
      (format nil "About ~a" *arc-name*)
      (format nil "A&bout ~a" *arc-name*)) :about-arc)
(defmeth arc :about-arc ()
"Method args: ()
Displays an about box."
    (message-dialog (format nil
"~a ~a Copyright (c) ~a R. Dennis Cook and
Sanford Weisberg.  Type (help 'copyright) to see the
license agreement.   See the file COPYING for the license
agreement for XLISP and XLISP-STAT." 
    *arc-name* *arc-version* "1994-99")))

(rc-menu-item 'help-menu-item (ms "Help..." "Help..." "&Help") :help-dialog)
(defmeth arc :help-dialog () 
"Method args: ()
Opens the help dialog."
(help-dialog))


(rc-menu-item 'settings-menu-item (ms "Settings..." "Settings..." "&Settings")
  :settings-namelist)

(rc-menu-item 'quit-menu-item (ms "Quit" "Quit" "E&xit") :polite-quit)
(defmeth arc :polite-quit ()
"Method args: ()
Opens quit dialog."
  (if (equal *computer-type* 'windows) (msw-exit)
      (if (ok-or-cancel-dialog "Are you sure you want to quit?
All information will be lost!")
         (exit))))

;;;;
;;;; help-dialog
;;;; Is a dialog-box interface to the standard xlispstat help file.  
;;;; Includes an interface to apropos.  The code to see how the help file
;;;; is accessed is borrowed from the standard help.lsp file.  
;;;; apropos has been disabled.
;;;;

(defmeth arc :help (arg) 
"Method args: arg
Finds the documentation for the STRING arg, if it exists."
  (find-documentation arg))

(defun help-dialog (&optional (initial-text "" set))
  (let* ((title (send text-item-proto :new 
                 (format nil "~a help facility" *arc-name*)))
         (instructions (send text-item-proto :new 
                         "Help for (type the name of a function or method)"))
         (get-help-item (send edit-text-item-proto :new 
                         (if set initial-text "") 
                         :text-length (apply #'ms *text-width-long*)))
;        (choose-star-help (send toggle-item-proto 
;            :new "Find all functions that include this text (apropos)" 
;            :value nil))
         (cancel-button (send modal-button-proto :new "Cancel"))
         (ok-button (send modal-button-proto  :new "OK" 
           :action #'(lambda () (list (send get-help-item :text)))))
         (help-dialog (send modal-dialog-proto
             :new (list title instructions get-help-item 
;                  choose-star-help
                   (list ok-button cancel-button))
             :default-button ok-button))
         (from-dialog (send help-dialog :modal-dialog)))
    (when from-dialog
     (when (null (equal (string-trim " " (first from-dialog)) ""))
#|
     (if (send choose-star-help :value) 
         (let* ((sl (string-to-symbol (first from-dialog)))
           (help-list (sort-data (mapcar #'string (apropos-list sl))))
           (help-dialog2 (let* ((label (send text-item-proto 
                         :new "Help may be available on: "))
                 (list-of-help  (send list-item-proto :new help-list))
                 (cancel-button (send modal-button-proto :new "Cancel"))
                 (ok-button (send modal-button-proto :new 
                        "Get more help on selected item"
                        :action 
                        #'(lambda () 
                            (send list-of-help :selection))))
                 (scroll-help-dialog (send modal-dialog-proto :new 
                            (list label list-of-help
                             (list ok-button 
                              cancel-button))
                            :default-button ok-button))
                 (from-scroll-dialog (send scroll-help-dialog 
                            :modal-dialog)))
                 (when from-scroll-dialog
                  (help-dialog (select help-list from-scroll-dialog)))))))
|#
       (message-dialog
        (let* ((dox (find-documentation (string-to-symbol (first from-dialog))))
          (real-dox (if (not dox) (format nil
"Sorry, no function or method ~a has
been found.  Check your spelling or use the 
apropos option in the help dialog." (first from-dialog))
                  (reline dox))))
          real-dox)))))) ;)


(defun reline (in &optional (char 80) out)
"Function args: (in &optional (char 80))
Inserts line feeds so no line is longer than char characters."
 (cond  ((< (length in) char)
         (if out (format nil "~a~%~a" out in) 
                 (format nil "~a" in)))
        (t (let* ((line (select in (iseq char)))
                  (newline (position #\Newline line))
                  (space (position #\Space line :from-end t))
                  (end (cond (newline newline) 
                             (space space) 
                             (t (- 1 char)))))
           (reline (select in (iseq (1+ end) (- (length in) 1)))
            char
            (if out (format nil "~a~%~a" out (select in (iseq end)))
                    (format nil "~a" (select in (iseq end)))
      ))))))

(defun string-to-symbol (strng)
   (eval (read (make-string-input-stream (format nil "'~a" strng)))))

(defparameter *protos* (list
;arc, dataset, datalist
  'arc 'dataset-proto 'datalist-proto 
;models
  'regression-model-proto 'glim-proto 'nonlin-model-proto 
  'inverse-regression-model-proto 'greg-proto 
;graphs
  'graph-proto 'histogram-proto 'scatterplot-proto 'scatmat-proto 
  'spin-proto 'boxplot-proto 'smoother-proto ))

(defun find-documentation (what-on)
"Function args (what-on)
Returns the documentation string for what if what-on is a function or a method
for one of the protos in the parameter *protos*."
 (let* ((dox (documentation what-on 'function)))
  (cond (dox (format nil "~a function:~%~a" 
               (string-capitalize (string-downcase what-on)) dox))
        (t (find-proto-documentation what-on)))))

(defun find-proto-documentation (what-on &optional (protos *protos*))
"Function args: (what-on &optional (proto *protos*))
If what-on is a method for one of the protos in the list, return the
documentation; otherwise, return nil."
 (cond ((null protos) nil)
       ((send (eval (first protos)) :has-method what-on :own t)
        (let ((dox (send (eval (first protos)) :documentation what-on)))
        (if dox
         (format nil "~a is a ~a method:~%~a" (string-downcase what-on)
           (string-capitalize (string-downcase (first protos))) dox)
         (format nil "~a is a ~a method,~%but no help is available."
           (string-downcase what-on)
           (string-capitalize (string-downcase (first protos)))))))
       (t (find-proto-documentation what-on (rest protos)))))

(send arc :menu)

;;
;;  BUGS
;;  1.  Double click to replace option-click
;;

(defproto arc-setting-proto '(symbol text value default) () )

(defmeth arc-setting-proto :isnew (symbol value text default)
"Method args: (symbol value text default)
Creates an instance of a arc-setting-proto with symbol symbol,
value value, text text and default default."
 (send self :symbol symbol)
 (send self :value value)
 (send self :default default)
 (send self :text text))

(defmeth arc-setting-proto :symbol (&optional (new nil set))
"Method args: (&optional (new nil set))
Returns/sets the symbol."
 (when set (setf (slot-value 'symbol) new))
 (slot-value 'symbol))

(defmeth arc-setting-proto :default (&optional (new nil set))
"Method args: (&optional (new nil set))
Returns/sets the default value of the setting."
 (when set (setf (slot-value 'default) new))
 (slot-value 'default))

(defmeth arc-setting-proto :text (&optional (new nil set))
"Method args: (&optional (new nil set))
Returns/sets the text."
 (when set (setf (slot-value 'text) new))
 (slot-value 'text))

(defmeth arc-setting-proto :value (&optional (new nil set))
"Method args: (&optional (new nil set))
Returns/sets the value."
 (when set (setf (slot-value 'value) new)
           (set (intern (string-upcase (slot-value 'symbol))) new))
 (slot-value 'value))

(defmeth arc-setting-proto :save ()
"Method args: ()
Returns code that when funcalled recreates the object."
 `(arc-setting ',(send self :symbol) 
                 ',(send self :value) 
                  ,(send self :text)
                 ',(send self :default)))

;(send arc :add-slot 'settings nil)

(defmeth arc :has-setting (symbol)
"Method args: (symbol)
Returns the position of symbol in the list of settings."
 (let ((pars (slot-value 'settings)))
  (position symbol (mapcar #'(lambda (a) (send a :symbol)) pars)
    :test #'equal)))

(defmeth arc :add-setting (p)
"Method args: (p)
Adds a setting p to the list of settings.  p must be an instance of a
arc-setting-proto."
 (let ((old (send self :has-setting (send p :symbol))))
  (if old (send self :delete-setting  (select (slot-value 'settings) old)))
  (setf (slot-value 'settings) 
       (append (slot-value 'settings) (list p)))))

(defmeth arc :delete-setting (p)
"Method args: (p)
Removes setting p from the list of settings."
 (send self :slot-value 'settings
  (remove p (slot-value 'settings) :test #'equal)))

(defmeth arc :settings () 
"Method args: ()
Returns a list of the settings."
(slot-value 'settings))

(defmeth arc :find-setting (symbol)
"Method args: (name)
Returns the arc-setting-proto with name name, or nil if no such object."
 (let* ((settings (slot-value 'settings))
        (symbols (mapcar #'(lambda (a) (send a :symbol)) settings))
        (pos (position symbol symbols :test #'equal)))
  (if pos (select settings pos) nil)))

(defmeth arc :save-settings (&optional (file "settings"))
"Method args: (&optional (file \"settings\"))
Save the current settings in the file file."
  (sysbeep)
  (when (ok-or-cancel-dialog "Over-write existing settings.lsp?")
        (let* ((len (length file))
               (f (if (equal (string-upcase (substr file (- len 4) 4)) ".LSP")
                      file (format nil "~a.lsp" file )))
               (f2 (format nil "~a~a~a" (get-working-directory) 
                         (ms "" "/" "\\") f))
               (f1 (open f2 :direction :output)))
          (format f1 "
;;;
;;;  Arc Settings File --- Created on ~a
;;;~%" (time-stamp))
         (mapcar #'(lambda (s) 
            (let ((a (send s :save)))
             (format f1 "~%(arc-setting ")
             (prin1 (second a) f1)
             (format f1 "~%")
             (prin1 (third a) f1)
             (format f1 "~%")
             (prin1 (fourth a) f1)
             (format f1 "~%")
             (prin1 (fifth a) f1)
             (format f1 "~%)~%"))) (send self :settings))
         (close f1))))


(defmeth arc :read-settings (&optional (file "settings.lsp")) 
"Method args: (&optional (file \"settings.lsp\"))
Reads the settings file named file."
(load file))

(defmeth arc :setting-dialog (symbol)
"Method args: (symbol)
Symbol is the name of an instance of an arc-setting-proto.  Opens a
dialog to modify symbol."
 (let* ((setting (send self :find-setting symbol))
        (ans (setting-dialog symbol 
                     (send setting :text) (send setting :value)
                     (send setting :default))))
    (when ans
       (send setting :value (first ans))
       (when (second ans) (send self :save-settings)))))

(defun setting-dialog (symbol text value default)
"Function args: (symbol text value)
Opens a dialog to modify a setting."
 (flet ((f (a) (format nil "~a" a)))
  (let* ((s1 (send text-item-proto :new (format nil "Setting = ~a" symbol)))
         (s2 (send text-item-proto :new text))
         (s3 (send text-item-proto :new (split-string
                (format nil "Current value = ~a" value))))
         (s5 (send text-item-proto :new (split-string
                (format nil "Default = ~a" default))))
         (s4 (send text-item-proto :new "New value = "))
         (t1 (send edit-text-item-proto :new ""
                 :text-length (apply #'ms *text-width-medium*)))
         (permanent (send toggle-item-proto :new 
                      "Make change and save new settings file" :value nil))
         (get-answer #'(lambda ()  
            (let* ((txt (string-trim " " (send t1 :text)))
                   (arg (send permanent :value))
                   (val (send t1 :value)) 
                   (ans (cond ((equal txt "") 'empty)
                              ((equal val t) val)
                              ((numberp val) val)
                              ((listp val) val)
                              (t val)))) ; changed txt to val 971002
              (if (equal ans 'empty) nil (list ans arg)))))
         (cancel (send modal-button-proto :new "Cancel"))
         (done (send modal-button-proto :new "OK" :action 
                     #'(lambda () (funcall get-answer))))
         (deflt (send modal-button-proto :new "Use Default" :action 
                     #'(lambda () (list default (send permanent :value)))))
         (dialog (send modal-dialog-proto :new
                       (list s1
                             s2 
                             s3
                             (list s4 t1) s5 permanent
                             (list done deflt cancel)))))
        (send dialog :default-button done)
        (send dialog :modal-dialog) )))

(defmeth arc :settings-namelist ()
"Method args: ()
Opens a namelist with all the settings."
  (let* ((settings (send self :settings))
         (symbols (qsort (mapcar #'(lambda (s) (send s :symbol)) settings)))
         (n (send name-list-proto :new 0 :show nil))
         (rc self))
    (send n :title (format nil "Settings List"))
    (send n :add-points (length symbols)) 
    (send n :point-label (iseq (length symbols)) 
       (mapcar #'(lambda (s) (format nil "~a" s)) symbols))
    (send n :menu (send self :settings-menu n))
    (defmeth n :do-click (x y m1 m2)
      (call-next-method x y m1 m2)
      (when m2
       (let ((sel (send self :points-selected)))
         (when sel (mapcar #'(lambda (sym) 
             (send rc :setting-dialog (select symbols sym))) sel)))))
    (send n :show-window)
    n))

(defmeth arc :settings-menu (n)
  (let* ((m (send menu-proto :new "Settings"))
         (i1 (send menu-item-proto :new "Update selection"
               :action #'(lambda () (send self :update-setting n))))
         (i2 (send menu-item-proto :new "Change all settings to defaults"
               :action #'(lambda () (send self :default-settings))))
         (i3 (send menu-item-proto :new "Save settings file"
               :action #'(lambda () (send self :save-settings))))
         (i4 (send menu-item-proto :new "Help with settings" 
               :action #'(lambda () (send self :setting-help)))))
   (send m :append-items i1 i2 i3 i4)
   (defmeth i1 :update ()
     (send self :enabled (send n :selection)))
   m))

(defmeth arc :setting-help ()
  (big-message-dialog 
"This is a list of settings that Arc uses to control many features in
the program.  You can change the values of these settings by clicking 
on the name of the setting, and then selecting \"Update setting\"
from the list's menu; equivalently, you can option-click on the 
name."
"Changes you make will be for this session only unless you save to a
file.  You can do this either from the dialog or from the list's menu."))
 

(defmeth arc :update-setting (n)
  (let ((sel (send n :points-selected))
        (symbols (qsort (mapcar #'(lambda (s) (send s :symbol)) 
             (send self :settings)))))
    (when sel (mapcar #'(lambda (sym)
       (send self :setting-dialog (select symbols sym))) sel))))

(defmeth arc :default-settings ()
 (sysbeep)
 (when (ok-or-cancel-dialog 
"Are you sure you want to return all 
settings to their default values?")
     (mapcar #'(lambda (s) (send s :value (send s :default)))
               (send self :settings))))

(defun arc-setting (symbol value text &optional (default value))
 (unless (send arc :has-setting symbol))
         (send arc :add-setting
                 (send arc-setting-proto :new symbol value text default)))

(defun rcode-setting (&rest args) (apply #'arc-setting args))

